@@ -40,6 +40,57 @@ int mem_poison_byte = -1;
 static int mem_fail_rate = 0;
 #endif
 
+#if defined(HA_HAVE_MALLOC_TRIM)
+static int using_libc_allocator = 0;
+
+/* ask the allocator to trim memory pools */
+static void trim_all_pools(void)
+{
+	if (using_libc_allocator)
+		malloc_trim(0);
+}
+
+/* check if we're using the same allocator as the one that provides
+ * malloc_trim() and mallinfo(). The principle is that on glibc, both
+ * malloc_trim() and mallinfo() are provided, and using mallinfo() we
+ * can check if malloc() is performed through glibc or any other one
+ * the executable was linked against (e.g. jemalloc).
+ */
+static void detect_allocator(void)
+{
+#ifdef HA_HAVE_MALLINFO2
+	struct mallinfo2 mi1, mi2;
+#else
+	struct mallinfo mi1, mi2;
+#endif
+	void *ptr;
+
+#ifdef HA_HAVE_MALLINFO2
+	mi1 = mallinfo2();
+#else
+	mi1 = mallinfo();
+#endif
+	ptr = DISGUISE(malloc(1));
+#ifdef HA_HAVE_MALLINFO2
+	mi2 = mallinfo2();
+#else
+	mi2 = mallinfo();
+#endif
+	free(DISGUISE(ptr));
+
+	using_libc_allocator = !!memcmp(&mi1, &mi2, sizeof(mi1));
+}
+#else
+
+static void trim_all_pools(void)
+{
+}
+
+static void detect_allocator(void)
+{
+}
+#endif
+
 /* Try to find an existing shared pool with the same characteristics and
  * returns it, otherwise creates this one. NULL is returned if no memory
  * is available for a new creation. Two flags are supported :
@@ -147,6 +198,14 @@ void *pool_get_from_os(struct pool_head *pool)
  */
 void pool_put_to_os(struct pool_head *pool, void *ptr)
 {
+#ifdef DEBUG_UAF
+	/* This object will be released for real in order to detect a use after
+	 * free. We also force a write to the area to ensure we crash on double
+	 * free or free of a const area.
+	 */
+	*(uint32_t *)ptr = 0xDEADADD4;
+#endif /* DEBUG_UAF */
+
 	pool_free_area(ptr, pool->size + POOL_EXTRA);
 	_HA_ATOMIC_DEC(&pool->allocated);
 }
@@ -271,79 +330,44 @@ void pool_flush(struct pool_head *pool)
 /* This function might ask the malloc library to trim its buffers. */
 void pool_gc(struct pool_head *pool_ctx)
 {
-#if defined(HA_HAVE_MALLOC_TRIM)
-	malloc_trim(0);
-#endif
+	trim_all_pools();
 }
 
-#elif defined(CONFIG_HAP_LOCKLESS_POOLS)
+#else /* CONFIG_HAP_NO_GLOBAL_POOLS */
+
+#if defined(CONFIG_HAP_LOCKLESS_POOLS)
 
 /*
  * This function frees whatever can be freed in pool <pool>.
  */
 void pool_flush(struct pool_head *pool)
 {
-	struct pool_free_list cmp, new;
-	void **next, *temp;
+	void *next, *temp;
 
 	if (!pool)
 		return;
-	HA_SPIN_LOCK(POOL_LOCK, &pool->lock);
+
+	/* The loop below atomically detaches the head of the free list and
+	 * replaces it with a NULL. Then the list can be released.
+	 */
+	next = pool->free_list;
 	do {
-		cmp.free_list = pool->free_list;
-		cmp.seq = pool->seq;
-		new.free_list = NULL;
-		new.seq = cmp.seq + 1;
-	} while (!_HA_ATOMIC_DWCAS(&pool->free_list, &cmp, &new));
+		while (unlikely(next == POOL_BUSY)) {
+			__ha_cpu_relax();
+			next = _HA_ATOMIC_LOAD(&pool->free_list);
+		}
+		if (next == NULL)
+			return;
+	} while (unlikely((next = _HA_ATOMIC_XCHG(&pool->free_list, POOL_BUSY)) == POOL_BUSY));
+	_HA_ATOMIC_STORE(&pool->free_list, NULL);
 	__ha_barrier_atomic_store();
-	HA_SPIN_UNLOCK(POOL_LOCK, &pool->lock);
-	next = cmp.free_list;
+
 	while (next) {
 		temp = next;
 		next = *POOL_LINK(pool, temp);
 		pool_put_to_os(pool, temp);
 	}
-	pool->free_list = next;
-	/* here, we should have pool->allocate == pool->used */
-}
-
-/*
- * This function frees whatever can be freed in all pools, but respecting
- * the minimum thresholds imposed by owners. It makes sure to be alone to
- * run by using thread_isolate(). <pool_ctx> is unused.
- */
-void pool_gc(struct pool_head *pool_ctx)
-{
-	struct pool_head *entry;
-	int isolated = thread_isolated();
-
-	if (!isolated)
-		thread_isolate();
-
-	list_for_each_entry(entry, &pools, list) {
-		while ((int)((volatile int)entry->allocated - (volatile int)entry->used) > (int)entry->minavail) {
-			struct pool_free_list cmp, new;
-
-			cmp.seq = entry->seq;
-			__ha_barrier_load();
-			cmp.free_list = entry->free_list;
-			__ha_barrier_load();
-			if (cmp.free_list == NULL)
-				break;
-			new.free_list = *POOL_LINK(entry, cmp.free_list);
-			new.seq = cmp.seq + 1;
-			if (HA_ATOMIC_DWCAS(&entry->free_list, &cmp, &new) == 0)
-				continue;
-			pool_put_to_os(entry, cmp.free_list);
-		}
-	}
-
-	if (!isolated)
-		thread_release();
-
-#if defined(HA_HAVE_MALLOC_TRIM)
-	malloc_trim(0);
-#endif
+	/* here, we should have pool->allocated == pool->used */
 }
 
 #else /* CONFIG_HAP_LOCKLESS_POOLS */
@@ -353,25 +377,26 @@ void pool_gc(struct pool_head *pool_ctx)
  */
 void pool_flush(struct pool_head *pool)
 {
-	void *temp;
+	void *temp, **next;
 
 	if (!pool)
 		return;
 
-	while (1) {
-		HA_SPIN_LOCK(POOL_LOCK, &pool->lock);
-		temp = pool->free_list;
-		if (!temp) {
-			HA_SPIN_UNLOCK(POOL_LOCK, &pool->lock);
-			break;
-		}
-		pool->free_list = *POOL_LINK(pool, temp);
-		HA_SPIN_UNLOCK(POOL_LOCK, &pool->lock);
+	HA_SPIN_LOCK(POOL_LOCK, &pool->lock);
+	next = pool->free_list;
+	pool->free_list = NULL;
+	HA_SPIN_UNLOCK(POOL_LOCK, &pool->lock);
+
+	while (next) {
+		temp = next;
+		next = *POOL_LINK(pool, temp);
 		pool_put_to_os(pool, temp);
 	}
 	/* here, we should have pool->allocated == pool->used */
 }
 
+#endif /* CONFIG_HAP_LOCKLESS_POOLS */
+
 /*
  * This function frees whatever can be freed in all pools, but respecting
  * the minimum thresholds imposed by owners. It makes sure to be alone to
@@ -396,10 +421,12 @@ void pool_gc(struct pool_head *pool_ctx)
 		}
 	}
 
+	trim_all_pools();
+
 	if (!isolated)
 		thread_release();
 }
-#endif /* CONFIG_HAP_LOCKLESS_POOLS */
+#endif /* CONFIG_HAP_NO_GLOBAL_POOLS */
 
 #else  /* CONFIG_HAP_POOLS */
 
@@ -411,9 +438,7 @@ void pool_flush(struct pool_head *pool)
 /* This function might ask the malloc library to trim its buffers. */
 void pool_gc(struct pool_head *pool_ctx)
 {
-#if defined(HA_HAVE_MALLOC_TRIM)
-	malloc_trim(0);
-#endif
+	trim_all_pools();
 }
 
 #endif /* CONFIG_HAP_POOLS */
@@ -564,6 +589,7 @@ static void init_pools()
 		LIST_INIT(&ha_thread_info[thr].pool_lru_head);
 	}
 #endif
+	detect_allocator();
 }
 
 INITCALL0(STG_PREPARE, init_pools);
